import os, subprocess, shutil, stat
from mama.parse_mamafile import parse_mamafile, update_mamafile_tag, update_cmakelists_tag
from mama.system import System, console, execute
from mama.util import is_dir_empty, has_tag_changed, write_text_to, read_lines_from, forward_slashes, back_slashes

######################################################################################

class Git:
    def __init__(self, url, branch, tag):
        if not url: raise RuntimeError("Git url must not be empty!")
        self.url = url
        self.branch = branch
        self.tag = tag
        self.dep = None
        self.missing_status = False
        self.url_changed = False
        self.tag_changed = False
        self.branch_changed = False
        self.commit_changed = False

    def run_git(self, git_command):
        cmd = f"cd {self.dep.src_dir} && git {git_command}"
        #console(cmd)
        execute(cmd)

    def current_commit(self): 
        cp = subprocess.run(['git','show','--oneline','-s'], stdout=subprocess.PIPE, cwd=self.dep.src_dir)
        return cp.stdout.decode('utf-8').rstrip()

    def save_status(self):
        status = f"{self.url}\n{self.tag}\n{self.branch}\n{self.current_commit()}\n"
        write_text_to(f"{self.dep.build_dir}/git_status", status)

    def check_status(self):
        lines = read_lines_from(f"{self.dep.build_dir}/git_status")
        if not lines:
            self.missing_status = True
            if not self.url: return False
            #console(f'check_status {self.url}: NO STATUS AT {self.dep.build_dir}/git_status')
            self.url_changed = True
            self.tag_changed = True
            self.branch_changed = True
            self.commit_changed = True
            return True
        self.url_changed = self.url != lines[0].rstrip()
        self.tag_changed = self.tag != lines[1].rstrip()
        self.branch_changed = self.branch != lines[2].rstrip()
        self.commit_changed = self.current_commit() != lines[3].rstrip()
        #console(f'check_status {self.url}: urlc={self.url_changed} tagc={self.tag_changed} brnc={self.branch_changed} cmtc={self.commit_changed}')
        return self.url_changed or self.tag_changed or self.branch_changed or self.commit_changed

    def branch_or_tag(self):
        if self.branch: return self.branch
        if self.tag: return self.tag
        return ''

    def checkout_current_branch(self):
        branch = self.branch_or_tag()
        if branch:
            if self.tag and self.tag_changed:
                self.run_git("reset --hard")
            self.run_git(f"checkout {branch}")

    def reclone_wipe(self):
        console(f'  - Target {self.dep.name: <16}   RECLONE WIPE')
        if os.path.exists(self.dep.dep_dir):
            if System.windows: # chmod everything to user so we can delete:
                for root, dirs, files in os.walk(self.dep.dep_dir):
                    for d in dirs:  os.chmod(os.path.join(root, d), stat.S_IWUSR)
                    for f in files: os.chmod(os.path.join(root, f), stat.S_IWUSR)
            shutil.rmtree(self.dep.dep_dir)

    def clone_or_pull(self, wiped=False):
        if is_dir_empty(self.dep.src_dir):
            if not wiped:
                console(f"  - Target {self.dep.name: <16}   CLONE because src is missing")
            branch = self.branch_or_tag()
            if branch: branch = f" --branch {self.branch_or_tag()}"
            execute(f"git clone --depth 1 {branch} {self.url} {self.dep.src_dir}")
            self.checkout_current_branch()
        else:
            self.checkout_current_branch()
            if not self.tag: # never pull a tag
                self.run_git("reset --hard")
                self.run_git("pull")

class BuildDependency:
    loaded_deps = dict()
    def __init__(self, name, config, target_class, workspace=None, src=None, git=None, is_root=False, mamafile=None):
        self.name       = name
        self.workspace  = workspace
        self.config     = config
        self.target     = None
        self.target_class = target_class
        self.mamafile     = mamafile
        self.should_rebuild = False
        self.nothing_to_build = False
        self.already_loaded = False
        self.already_executed  = False
        self.is_root = is_root # Root deps are always built
        self.children = []
        self.depends_on = []
        self.product_sources = []
        if not src and not git:
            raise RuntimeError(f'{name} src and git not configured. Specify at least one.')

        if git:
            self.git     = git
            git.dep      = self
            self.update_dep_dir()
            self.src_dir = forward_slashes(os.path.join(self.dep_dir, self.name))
            self.target  = None
        else:
            self.git     = None
            self.src_dir = forward_slashes(src)
            self.create_build_target()
            self.name = self.target.name
            self.update_dep_dir()
    
    
    @staticmethod
    def get(name, config, target_class, workspace=None, src=None, git=None, mamafile=None):
        if name in BuildDependency.loaded_deps:
            return BuildDependency.loaded_deps[name]
            
        dependency = BuildDependency(name, config, target_class, \
                        workspace=workspace, src=src, git=git, mamafile=mamafile)
        BuildDependency.loaded_deps[name] = dependency
        return dependency


    def update_dep_dir(self):
        dep_name = self.name
        if self.git:
            if self.git.branch: dep_name = f'{self.name}-{self.git.branch}'
            elif self.git.tag:  dep_name = f'{self.name}-{self.git.tag}'
        self.dep_dir   = forward_slashes(os.path.join(self.config.workspaces_root, self.workspace, dep_name))
        self.build_dir = forward_slashes(os.path.join(self.dep_dir, self.config.name()))
        

    def is_reconfigure_target(self):
        return self.config.configure and self.config.target == self.name

    def exported_libs_file(self):
        return self.build_dir + '/mama_exported_libs'

    def load_build_dependencies(self, target):
        for saved_dependency in read_lines_from(self.exported_libs_file()):
            target.add_build_dependency(saved_dependency)

    def save_exports_as_dependencies(self, exports):
        write_text_to(self.exported_libs_file(), '\n'.join(exports))


    def get_missing_build_dependency(self):
        for depfile in self.target.build_dependencies:
            if not os.path.getsize(depfile):
                return depfile
        return None

    def create_build_dir_if_needed(self):
        if not os.path.exists(self.build_dir): # check to avoid Access Denied errors
            os.makedirs(self.build_dir, exist_ok=True)

    ## @return True if dependency has changed
    def load(self):
        #console(f'LOAD {self.name}')
        git_changed = self.git_checkout()
        self.create_build_target()
        self.update_dep_dir()
        self.create_build_dir_if_needed()

        if git_changed:
            #console(f'SAVE GIT STATUS {self.target.name}')
            self.git.save_status() # save git status to avoid recloning

        target = self.target
        conf = self.config
        is_target = conf.target_matches(target.name)
        if conf.clean and is_target:
            self.clean()

        if not self.is_root:
            self.load_build_dependencies(target)
        target.dependencies() ## customization point for additional dependencies

        build = False
        if conf.build:
            build = True
            def reason(r): console(f'  - Target {target.name: <16}   BUILD [{r}]')

            if conf.target and not is_target:
                build = False # skip build if target doesn't match
            
            elif conf.clean and is_target: reason('cleaned target')
            elif self.is_root:             reason('root target')
            elif update_mamafile_tag(self.src_dir, self.build_dir):   reason(f'{target.name}/mamafile.py modified')
            elif update_cmakelists_tag(self.src_dir, self.build_dir): reason(f'{target.name}/CMakeLists.txt modified')
            elif git_changed:                   reason('git commit changed')
            elif self.is_reconfigure_target():  reason(f'configure target={target.name}')
            elif not target.build_dependencies: reason('no build dependencies')
            else:
                missing = self.get_missing_build_dependency()
                if missing: reason(f'{missing} does not exist')
                else:
                    console(f'  - Target {target.name: <16}   OK')
                    build = False

        self.already_loaded = True
        self.should_rebuild = build
        if build:
            self.create_build_dir_if_needed() # in case we just cleaned
        return build


    def create_build_target(self):
        if self.target:
            return

        project, buildTarget = parse_mamafile(self.config, self.src_dir, \
                                            self.target_class, mamafile=self.mamafile)
        if project and buildTarget:
            buildStatics = buildTarget.__dict__
            if not self.workspace:
                if   'workspace'        in buildStatics: self.workspace = buildStatics['workspace']
                elif 'local_workspace'  in buildStatics: self.workspace = buildStatics['local_workspace']
                elif 'global_workspace' in buildStatics: self.workspace = buildStatics['global_workspace']
                else:                                    self.workspace = 'build'
            if self.is_root:
                if   'workspace'        in buildStatics: self.config.global_workspace = False
                elif 'local_workspace'  in buildStatics: self.config.global_workspace = False
                elif 'global_workspace' in buildStatics: self.config.global_workspace = True
                if not self.config.global_workspace:
                    self.config.workspaces_root = self.src_dir
            self.target = buildTarget(name=project, config=self.config, dep=self)
        else:
            if not self.workspace:
                self.workspace = 'build'
            self.target = self.target_class(name=self.name, config=self.config, dep=self)


    def git_checkout(self):
        if not self.git:
            return False
        
        # if no update command, allow us to skip pulling by returning False
        changed = self.git.check_status()
        is_target = self.config.target_matches(self.name)
        update = self.git.commit_changed or (self.config.update and is_target)
        if not changed and not update:
            return False

        wiped = False
        should_wipe = self.git.url_changed and not self.git.missing_status
        if should_wipe or (is_target and self.config.reclone):
            self.git.reclone_wipe()
            wiped = True
        self.git.clone_or_pull(wiped)
        return True

    ## GIT
    def save_git_status(self):
        if self.git: self.git.save_status()

    ## Clean
    def clean(self):
        console(f'  - Target {self.name: <16}   CLEAN')

        if self.build_dir == '/' or not os.path.exists(self.build_dir):
            return
        
        self.target.clean() # Customization point
        shutil.rmtree(self.build_dir, ignore_errors=True)



import os, concurrent.futures
from .build_target import BuildTarget
from .build_dependency import BuildDependency
from .util import save_file_if_contents_changed
from .system import console

def get_mama_dependencies_cmake(root_dependency: BuildDependency, build:str):
    if not root_dependency.children:
        return ''
    return f'''# get MAMA_INCLUDES and MAMA_LIBS for this platform; verbose for CLion CMake parser
    include("{root_dependency.dep_dir}/{build}/mama-dependencies.cmake")'''

def save_generic_cmake(root_dependency: BuildDependency):
    outfile = f'{root_dependency.src_dir}/mama.cmake'
    # note: we save verbose include directives, because CLion has a hard time detecting macro paths
    text = f'''
# This file is auto-generated by mama build. Do not modify by hand!
if(ANDROID OR ANDROID_NDK)
    set(MAMA_BUILD "android")
    {get_mama_dependencies_cmake(root_dependency, 'android')}
elseif(WIN32)
    set(MAMA_BUILD "windows")
    {get_mama_dependencies_cmake(root_dependency, 'windows')}
elseif(APPLE)
  if(IOS_PLATFORM)
    set(IOS TRUE)
    set(MAMA_BUILD "ios")
    {get_mama_dependencies_cmake(root_dependency, 'ios')}
  else()
    set(MACOS TRUE)
    set(MAMA_BUILD "macos")
    {get_mama_dependencies_cmake(root_dependency, 'macos')}
  endif()
elseif(UNIX)
    set(LINUX TRUE)
    set(MAMA_BUILD "linux")
    {get_mama_dependencies_cmake(root_dependency, 'linux')}
else()
    message(FATAL_ERROR "mama build: Unsupported Platform!")
    set(MAMA_BUILD "???")
endif()
if(CMAKE_CXX_COMPILER_ID MATCHES "Clang")
    set(CLANG TRUE)
elseif(CMAKE_CXX_COMPILER_ID MATCHES "GNU")
    set(GCC TRUE)
endif()
'''
    save_file_if_contents_changed(outfile, text)


def get_cmake_path_list(paths):
    pathlist = '' 
    for path in paths: pathlist += f'\n    "{path}"'
    return pathlist


def save_dependencies_cmake(root_dependency: BuildDependency):
    outfile = f'{root_dependency.build_dir}/mama-dependencies.cmake'
    if not root_dependency.children:
        if os.path.exists(outfile):
            os.remove(outfile) # no more deps, get rid of the dependency file
        return
    text = '''
# This file is auto-generated by mama build. Do not modify by hand!
set(MAMA_INCLUDES "")
set(MAMA_LIBS     "")
'''
    for dep in root_dependency.children:
        includes  = get_cmake_path_list(dep.target.exported_includes)
        libraries = get_cmake_path_list(dep.target.exported_libs)
        text += f'''
# Package {dep.name}
set(MAMA_INCLUDES ${{MAMA_INCLUDES}} {includes})
set(MAMA_LIBS     ${{MAMA_LIBS}}     {libraries})
'''
    save_file_if_contents_changed(outfile, text)


def create_mama_cmake_includes(root_dependency: BuildDependency):
    save_dependencies_cmake(root_dependency)
    save_generic_cmake(root_dependency)


def load_child_dependencies(root_dependency: BuildDependency, parallel=False):
    changed = False
    if parallel:
        futures = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as e:
            for dep in root_dependency.children:
                futures.append(e.submit(load_dependency_chain, dep))
        for f in futures:
            changed |= f.result()
    else:
        for dep in root_dependency.children:
            changed |= load_dependency_chain(dep)
    return changed


def load_dependency_chain(dep: BuildDependency):
    if dep.already_loaded:
        return dep.should_rebuild

    changed = dep.load()
    changed |= load_child_dependencies(dep)
    return changed


def execute_task_chain(root_dependency: BuildDependency):
    for dep in root_dependency.children:
        execute_task_chain(dep)
    
    create_mama_cmake_includes(root_dependency)
    root_dependency.target.execute_tasks()

def find_dependency(root_dependency: BuildDependency, name) -> BuildDependency:
    if root_dependency.name == name:
        return root_dependency
    for dep in root_dependency.children:
        found = find_dependency(dep, name)
        if found: return found
    return None



import os, sys, multiprocessing, subprocess
from mama.system import System, console


def find_executable_from_system(name):
    finder = 'where' if System.windows else 'which'
    output = subprocess.run([finder, name], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).stdout.decode('utf-8')
    output = output.split('\n')[0].strip()
    return output if os.path.isfile(output) else ''

###
# Mama Build Configuration is created only once in the root project working directory
# This configuration is then passed down to dependencies
#
class BuildConfig:
    def __init__(self, args):
        self.build   = False
        self.clean   = False
        self.rebuild = False
        self.update  = False
        self.configure = False # re-run cmake configure
        self.reclone   = False
        self.test      = ''
        self.windows = False
        self.linux   = False
        self.macos   = False
        self.ios     = False
        self.android = False
        self.clang = True # prefer clang on linux
        self.gcc   = False   
        self.release = True
        self.debug   = False
        self.jobs    = multiprocessing.cpu_count()
        self.target  = None
        self.flags   = None
        self.open    = None
        self.fortran = ''
        self.ios_version   = '11.0'
        self.macos_version = '10.12'
        self.ninja_path = self.find_ninja_build()
        self.android_sdk_path = ''
        self.android_ndk_path = ''
        self.init_ndk_path()
        self.android_arch  = 'armeabi-v7a' # arm64-v8a
        self.android_tool  = 'arm-linux-androideabi-4.9' # aarch64-linux-android-4.9
        self.android_api   = 'android-24'
        self.android_ndk_stl = 'c++_shared' # LLVM libc++
        self.global_workspace = True
        self.workspaces_root = os.getenv('HOMEPATH') if System.windows else os.getenv('HOME')
        self.parse_args(args)
        self.check_platform()


    def set_platform(self, windows=False, linux=False, macos=False, ios=False, android=False):
        self.windows = windows
        self.linux   = linux
        self.macos   = macos
        self.ios     = ios
        self.android = android
        return True


    def is_platform_set(self):
        return self.windows or self.linux or self.macos or self.ios or self.android
    
    
    def check_platform(self):
        if not self.is_platform_set():
            self.set_platform(windows=System.windows, linux=System.linux, macos=System.macos)
            if not self.is_platform_set():
                raise RuntimeError(f'Unsupported platform {sys.platform}: Please specify platform!')
    
    
    def name(self):
        if self.windows: return 'windows'
        if self.linux:   return 'linux'
        if self.macos:   return 'macos'
        if self.ios:     return 'ios'
        if self.android: return 'android'
        return 'build'


    def set_build_config(self, release=False, debug=False):
        self.release = release
        self.debug   = debug
        return True


    def parse_args(self, args):
        for arg in args:
            if arg == 'build':     self.build   = True
            elif arg == 'clean':     self.clean   = True
            elif arg == 'rebuild':   self.rebuild = True
            elif arg == 'update':    self.update  = True
            elif arg == 'configure': self.configure = True
            elif arg == 'reclone':   self.reclone   = True
            elif arg == 'test':      self.test = ' ' # no test arguments
            elif arg == 'windows': self.set_platform(windows=True)
            elif arg == 'linux':   self.set_platform(linux=True)
            elif arg == 'macos':   self.set_platform(macos=True)
            elif arg == 'ios':     self.set_platform(ios=True)
            elif arg == 'android': self.set_platform(android=True)
            elif arg == 'clang':   
                self.gcc = False
                self.clang = True
            elif arg == 'gcc':
                self.gcc = True
                self.clang = False
            elif arg == 'fortran': self.fortran = self.find_default_fortran_compiler()
            elif arg == 'fortran=': self.fortran = arg[8:]
            elif arg == 'release': self.set_build_config(release=True)
            elif arg == 'debug':   self.set_build_config(debug=True)
            elif arg == 'open':    self.open = 'root'
            elif arg.startswith('open='):   self.open = arg[5:]
            elif arg.startswith('jobs='):   self.jobs = int(arg[5:])
            elif arg.startswith('target='): self.target = arg[7:]
            elif arg.startswith('test='):   self.add_test_arg(arg[5:])
            elif arg.startswith('flags='):
                self.flags = arg[6:]
            continue

    def add_test_arg(self, arg):
        if arg[0] == '"' and arg[-1] == '"':
            arg = arg[1:-1]
        if self.test: self.test += ' '
        self.test += arg

    def find_ninja_build(self):
        ninja_executables = [
            os.getenv('NINJA'), 
            find_executable_from_system('ninja'),
            '/Projects/ninja.exe'
        ]
        for ninja_exe in ninja_executables:        
            if ninja_exe and os.path.isfile(ninja_exe):
                #console(f'Found Ninja Build System: {ninja_exe}')
                return ninja_exe
        return ''


    def init_ndk_path(self):
        androidenv = os.getenv('ANDROID_HOME')
        paths = [androidenv] if androidenv else []
        if System.windows: paths += [f'{os.getenv("LOCALAPPDATA")}\\Android\\Sdk']
        elif System.linux: paths += ['/usr/bin/android-sdk', '/opt/android-sdk']
        elif System.macos: paths += [f'{os.getenv("HOME")}/Library/Android/sdk']
        ext = '.cmd' if System.windows else ''
        for sdk_path in paths:
            if os.path.exists(f'{sdk_path}/ndk-bundle/ndk-build{ext}'):
                self.android_sdk_path = sdk_path
                self.android_ndk_path = sdk_path  + '/ndk-bundle'
                #console(f'Found Android NDK: {self.ndk_path}')
                return
        return ''


    def find_default_fortran_compiler(self):
        paths = []
        if System.linux:
            paths += [find_executable_from_system('gfortran')]
        
        for fortran_path in paths:
            if fortran_path and os.path.exists(fortran_path):
                console(f'Found Fortran: {fortran_path}')
                return fortran_path
        return None


    def libname(self, library):
        if self.windows: return f'{library}.lib'
        else:            return f'lib{library}.a'


    def libext(self):
        return 'lib' if self.windows else 'a'


    def target_matches(self, target_name):
        return self.target == 'all' or self.target == target_name




import os.path, shutil
import pathlib, stat, time, subprocess, concurrent.futures
from mama.system import System, console
from mama.util import execute, save_file_if_contents_changed, glob_with_name_match, \
                    normalized_path, write_text_to
from mama.build_dependency import BuildDependency, Git
from mama.build_config import BuildConfig
from mama.cmake_configure import run_cmake_config, run_cmake_build, cmake_default_options, \
                            cmake_inject_env, cmake_buildsys_flags, cmake_generator
import mama.util as util

######################################################################################

##
# Describes a single configurable build target
# This is the main public interface for configuring a specific target
# For project-wide configuration, @see BuildConfig in self.config
#
# Customization points:
#
# class MyProject(mama.BuildTarget):
#     local_workspace = 'build'
#     
#
class BuildTarget:


    def __init__(self, name, config:BuildConfig, dep:BuildDependency):
        if config is None: raise RuntimeError('BuildTarget config argument must be set')
        if dep is None:    raise RuntimeError('BuildTarget dep argument must be set')
        self.config = config
        self.name = name
        self.dep = dep
        self.install_target   = 'install'
        self.cmake_ndk_toolchain = f'{config.android_ndk_path}/build/cmake/android.toolchain.cmake' if config.android_ndk_path else ''
        self.cmake_ios_toolchain = ''
        self.cmake_opts       = []
        self.cmake_cxxflags   = dict()
        self.cmake_ldflags    = dict()
        self.cmake_build_type = 'Debug' if config.debug else 'RelWithDebInfo'
        self.enable_exceptions = True
        self.enable_unix_make  = False
        self.enable_ninja_build = True and config.ninja_path # attempt to use Ninja
        self.enable_multiprocess_build = True
        self.build_dependencies = [] # dependency files
        self.exported_includes = [] # include folders to export from this target
        self.exported_libs     = [] # libs to export from this target
        self.windows = self.config.windows # convenient alias
        self.linux   = self.config.linux
        self.macos   = self.config.macos
        self.ios     = self.config.ios
        self.android = self.config.android

    def _get_full_path(self, path):
        if path and not os.path.isabs(path):
            if self.dep.mamafile: # if setting mamafile, then use mamafile folder:
                path = os.path.join(os.path.dirname(self.dep.mamafile), path)
            else:
                path = os.path.join(self.dep.src_dir, path)
            path = normalized_path(path)
        return path


    def _get_mamafile_path(self, name, mamafile):
        if mamafile:
            return self._get_full_path(mamafile)
        maybe_mamafile = self._get_full_path(f'mama/{name}.py')
        if os.path.exists(maybe_mamafile):
            return maybe_mamafile
        return mamafile


    ##
    #  Add a local dependency. This can be a git submodule or just some local folder
    #  which contains its own CMakeLists.txt
    #  Optionally you can override the default 'mamafile.py' with your own.
    #  
    #  If the local dependency folder does not contain a `mamafile.py`, you will have to
    #  provide your own relative or absolute mamafile path.
    # Ex:
    #   self.add_local('zlib', '3rdparty/zlib')
    #   self.add_local('zlib', '3rdparty/zlib', mamafile='mama/zlib.py')
    #
    def add_local(self, name, source_dir, mamafile=None):
        src      = self._get_full_path(source_dir)
        mamafile = self._get_mamafile_path(name, mamafile)
        dependency = BuildDependency.get(name, self.config, BuildTarget, \
                        workspace=self.dep.workspace, src=src, mamafile=mamafile)
        self.dep.children.append(dependency)


    ##
    #  Add a remote GIT dependency
    #  The dependency will be cloned and updated according to mamabuild.
    #  Use `mama update` to force update the git repositories.
    #
    #  If the remote GIT repository does not contain a `mamafile.py`, you will have to
    #  provide your own relative or absolute mamafile path.
    # Ex:
    #   self.add_git('ReCpp', 'https://github.com/RedFox20/ReCpp.git')
    #   self.add_git('ReCpp', 'https://github.com/RedFox20/ReCpp.git', git_branch='master')
    #   self.add_git('opencv', 'https://github.com/opencv/opencv.git', git_branch='3.4', mamafile='mama/opencv_cfg.py')
    #
    def add_git(self, name, git_url, git_branch='', git_tag='', mamafile=None):
        git = Git(git_url, git_branch, git_tag)
        mamafile = self._get_mamafile_path(name, mamafile)
        dependency = BuildDependency.get(name, self.config, BuildTarget, \
                        workspace=self.dep.workspace, git=git, mamafile=mamafile)
        self.dep.children.append(dependency)

    ##
    # Finds a dependency by name
    #
    def get_dependency(self, name):
        for dep in self.dep.children:
            if dep.name == name:
                return dep
        raise KeyError(f"BuildTarget {self.name} has no child dependency named '{name}'")


    ##
    #  Injects products from `src_dep` into `dst_dep` as CMake defines
    #  Name of defines is given via `include_path` and `libs` params
    #  `libfilters` does simple string matching; if nothing matches, the first export lib is chosen
    # Ex:
    #    self.inject_products('libpng', 'zlib', 'ZLIB_INCLUDE_DIR', 'ZLIB_LIBRARY', 'zlibstatic')
    #
    def inject_products(self, dst_dep, src_dep, include_path, libs, libfilters=None):
        dst_dep = self.get_dependency(dst_dep)
        src_dep = self.get_dependency(src_dep)
        dst_dep.depends_on.append(src_dep)
        dst_dep.product_sources.append( (src_dep, include_path, libs, libfilters) )

    ##
    #  Collects all results injected by `inject_products`
    #  Returns a list of injected defines:
    #  [ 'ZLIB_INCLUDE_DIR=path/to/zlib/include', 'ZLIB_LIBRARY=path/to/lib/zlib.a', ... ]
    #
    def get_product_defines(self):
        defines = []
        for source in self.dep.product_sources:
            srcdep = source[0]
            includes = srcdep.target.get_exported_includes()
            libraries = srcdep.target.get_exported_libs(source[3])
            #console(f'grabbing products: {srcdep.name}; includes={includes}; libraries={libraries}')
            defines.append(f'{source[1]}={includes}')
            defines.append(f'{source[2]}={libraries}')
        return defines

    def get_exported_includes(self):
        return ';'.join(self.exported_includes) if self.exported_includes else ''

    def get_exported_libs(self, libfilters):
        #console(f'libfilters={libfilters}')
        if not self.exported_libs: return
        if not libfilters: return ';'.join(self.exported_libs)
        libs = []
        for lib in self.exported_libs:
            if libfilters in lib: libs.append(lib)
        if not libs: libs.append(self.exported_libs[0])
        return ';'.join(libs)


    ##
    # Gets target products as a tuple: (include_paths=[], libs=[])
    # Ex:
    #  self.get_target_products('zlib')
    def get_target_products(self, target_name):
        dep = self.get_dependency(target_name)
        target:BuildTarget = dep.target
        return (target.get_exported_includes(), target.get_exported_libs())


    ## 
    # Manually add a build dependency to prevent unnecessary rebuilds
    # Normally the build dependency is detected from the packaged libraries.
    # 
    # if the dependency file does not exist, then the project will be rebuilt
    #
    # if your project has no build dependencies, it will always be rebuilt, so make sure
    # to add_build_dependency or export_lib
    #
    def add_build_dependency(self, all=None, windows=None, linux=None, macos=None, ios=None, android=None):
        dependency = all if all else self.select(windows, linux, macos, ios, android)
        if dependency:
            dependency = normalized_path(os.path.join(self.dep.build_dir, dependency))
            self.build_dependencies.append(dependency)
            #console(f'    {self.name}.build_dependencies += {dependency}')


    ##
    # CUSTOM PACKAGE INCLUDES (if self.default_package() is insufficient)
    # 
    # Export include path relative to source directory
    #  OR if build_dir=True, then relative to build directory
    # Ex:
    #   self.export_include('./include') # MyRepo/include
    #   self.export_include('installed/MyLib/include', build_dir=True) # CMake installed includes in build dir
    #
    def export_include(self, include_path, build_dir=False):
        root = self.dep.build_dir if build_dir else self.dep.src_dir
        include_path = normalized_path(os.path.join(root, include_path))
        #console(f'export_include={include_path}')
        if os.path.exists(include_path):
            if not include_path in self.exported_includes:
                self.exported_includes.append(include_path)
            return True
        return False

    ##
    # CUSTOM PACKAGE INCLUDES (if self.default_package() is insufficient)
    #
    # Export include paths relative to source directory
    #  OR if build_dir=True, then relative to build directory
    # Ex:
    #   self.export_includes(['include', 'src/moreincludes'])
    #   self.export_includes(['installed/include', 'installed/src/moreincludes'], build_dir=True)
    def export_includes(self, include_paths=[''], build_dir=False):
        self.exported_includes = []
        for include_path in include_paths:
            self.export_include(include_path, build_dir)


    ##
    # CUSTOM PACKAGE LIBS (if self.default_package() is insufficient)
    #
    # Export a specific lib relative to build directory
    #  OR if src_dir=True, then relative to source directory
    # Ex:
    #   self.export_lib('mylib.a')  # from build dir
    #   self.export_lib('lib/mylib.a', src_dir=True)  # from project source dir
    def export_lib(self, relative_path, src_dir=False):
        root = self.dep.src_dir if src_dir else self.dep.build_dir
        path = normalized_path(os.path.join(root, relative_path))
        if os.path.exists(path):
            self.exported_libs.append(path)
            self._remove_duplicate_export_libs()


    ##
    # CUSTOM PACKAGE LIBS (if self.default_package() is insufficient)
    #
    # Export several libs relative to build directory using EXTENSION MATCHING
    #  OR if src_dir=True, then relative to source directory
    #
    # Ex:
    #   self.export_libs()  # gather any .lib or .a from build dir
    #   self.export_libs('.', ['.dll', '.so'])   # gather any .dll or .so from build dir
    #   self.export_libs('lib', src_dir=True)  # export everything from project/lib directory
    #   self.export_libs('external/lib')  # gather specific static libs from build dir
    # 
    def export_libs(self, path = '.', pattern_substrings = ['.lib', '.a'], src_dir=False):
        root = self.dep.src_dir if src_dir else self.dep.build_dir
        path = os.path.join(root, path)
        self.exported_libs = glob_with_name_match(path, pattern_substrings)
        self._remove_duplicate_export_libs()
        return len(self.exported_libs) > 0


    def _remove_duplicate_export_libs(self):
        unique = dict()
        for lib in self.exported_libs:
            unique[os.path.basename(lib)] = lib
        self.exported_libs = list(unique.values())


    ##
    # Injects default platform and target specific environment variables
    # This can be used for custom build step
    # 
    def inject_env(self):
        cmake_inject_env(self)

    def _add_dict_flag(self, dest:dict, flag):
        if not flag: return
        if ' ' in flag:
            for subflag in flag.split(' '):
                self._add_dict_flag(dest, subflag)
        elif '=' in flag:
            key, value = flag.split('=', 1)
            dest[key] = value
        elif ':' in flag:
            key, value = flag.split(':', 1)
            dest[key] = value
        else:
            dest[flag] = ''

    ##
    #  Adds C / C++ flags for compilation step
    #  Supports many different usages: strings, list of strings, kwargs, or space separate string
    # Ex:
    #   self.add_cxx_flags('-Wall')
    #   self.add_cxx_flags(['-Wall', '-std=c++17'])
    #   self.add_cxx_flags('-Wall', '-std=c++17')
    #   self.add_cxx_flags('-Wall -std=c++17')
    def add_cxx_flags(self, *flags):
        for flag in flags:
            if isinstance(flag, list): self.add_cxx_flags(*flag)
            else: self._add_dict_flag(self.cmake_cxxflags, flag)
    
    ##
    #  Adds flags for linker step; No platform checking is done
    #  Supports many different usages: strings, list of strings, kwargs, or space separate string
    # Ex:
    #   self.add_ld_flags('-rdynamic')
    #   self.add_ld_flags(['-rdynamic', '-s'])
    #   self.add_ld_flags('-rdynamic', '-s')
    #   self.add_ld_flags('-rdynamic -s')
    def add_ld_flags(self, *flags):
        for flag in flags:
            if isinstance(flag, list): self.add_ld_flags(*flag)
            else: self._add_dict_flag(self.cmake_ldflags, flag)

    ##
    #  Adds C / C++ flags flags depending on configuration platform
    #  Supports many different usages: strings, list of strings, kwargs, or space separate string
    # Ex:
    #   self.add_cxx_flags('-Wall')
    #   self.add_cxx_flags(['-Wall', '-std=c++17'])
    #   self.add_cxx_flags('-Wall', '-std=c++17')
    #   self.add_cxx_flags('-Wall -std=c++17')
    #
    def add_platform_cxx_flags(self, windows=None, linux=None, macos=None, ios=None, android=None):
        flags = self.select(windows, linux, macos, ios, android)
        if flags: self.add_cxx_flags(flags)

    ##
    #  Adds linker flags depending on configuration platform
    #  Supports many different usages: strings, list of strings, or space separate string
    # Ex:
    #   self.add_platform_ld_flags(windows='/LTCG', ios=['-lobjc', '-rdynamic'], linux='-rdynamic -s')
    #
    def add_platform_ld_flags(self, windows=None, linux=None, macos=None, ios=None, android=None):
        flags = self.select(windows, linux, macos, ios, android)
        if flags: self.add_ld_flags(flags)


    ## 
    #  Main method for configuring CMake options
    # Ex:
    #   self.add_cmake_options('ZLIB_STATIC=TRUE', 'NO_GUI=1')
    #   self.add_cmake_options(['ZLIB_STATIC=TRUE', 'NO_GUI=1'])
    #
    def add_cmake_options(self, *options):
        for option in options:
            if isinstance(option, list): self.cmake_opts += option
            else:                        self.cmake_opts.append(option)

    
    ##
    #  Selectively applies CMake options depending on configuration platform
    # Ex:
    #  self.add_platform_options(windows='ZLIB_STATIC=TRUE')
    #
    def add_platform_options(self, windows=None, linux=None, macos=None, ios=None, android=None):
        defines = self.select(windows, linux, macos, ios, android)
        if defines: self.cmake_opts += defines


    def select(self, windows, linux, macos, ios, android):
        if   self.config.windows and windows: return windows
        elif self.config.linux   and linux:   return linux
        elif self.config.macos   and macos:   return macos
        elif self.config.ios     and ios:     return ios
        elif self.config.android and android: return android
        return None

    ##
    # Enable a specific C++ standard
    def enable_cxx20(self):
        self.cmake_cxxflags['/std' if self.config.windows else '-std'] = 'c++latest' if self.config.windows else 'c++2a'
    def enable_cxx17(self):
        self.cmake_cxxflags['/std' if self.config.windows else '-std'] = 'c++17'
    def enable_cxx14(self):
        self.cmake_cxxflags['/std' if self.config.windows else '-std'] = 'c++14'
    def enable_cxx11(self):
        self.cmake_cxxflags['/std' if self.config.windows else '-std'] = 'c++11'


    ##
    # Utility for copying files within the build directory
    #
    def copy_built_file(self, builtFile, copyToFolder):
        src = f'{self.dep.build_dir}/{builtFile}'
        dst = f'{self.dep.build_dir}/{copyToFolder}'
        shutil.copy(src, dst)


    def download_file(self, remote_url, local_dir, force=False):
        util.download_file(remote_url, local_dir, force)


    def download_and_unzip(self, remote_zip, extract_dir):
        util.download_and_unzip(remote_zip, extract_dir)
        

    ##
    # Call this to completely skip the build step every time
    # and instead 
    def nothing_to_build(self):
        self.dep.nothing_to_build = True
        self.dep.should_rebuild = False


    # Run a command in the build folder. Can be used for any custom commands or custom build systems
    # Ex:
    #  self.run('./configure')
    #  self.run('make release -j7')
    #
    def run(self, command):
        execute(f'cd {self.dep.build_dir} && {command}', echo=True)

    
    # Run a command with gdb in the build folder
    def gdb(self, command):
        if self.config.android or self.config.ios:
            return # nothing to run
        
        build_dir = self.dep.build_dir
        if self.config.windows:
            build_dir = f'{build_dir}/{self.cmake_build_type}'
            gdb = command.lstrip('./') # no gdb on windows
        else: # linux, macos
            gdb = f'gdb -batch -return-child-result -ex=r -ex=bt --args {command}'
        execute(f'cd {build_dir} && {gdb}', echo=True)

    ########## Customization Points ###########


    ###
    # Add any dependencies in this step
    #   self.add_local(...)
    #   self.add_remote(...)
    #
    def dependencies(self):
        pass


    ###
    # Perform any pre-build steps here
    def configure(self):
        pass


    ###
    # Build this target. By default it uses CMake build
    def build(self):
        self.cmake_build()


    ### 
    # Perform any pre-clean steps here
    def clean(self):
        pass


    ###
    # Perform custom install steps here. By default it uses CMake install
    def install(self):
        self.cmake_install()


    ###
    # Perform any post-build steps to package the products.
    # If no headers or libs are exported, then `default_package()` will be run instead
    #
    # Every library should at least export some headers
    # Ex:
    #   def package(self):
    #       self.export_libs('.', ['.lib', '.a']) # export any .lib or .a from build folder
    #       self.export_includes(['AGL']) # export AGL as include from source folder
    #
    def package(self):
        pass
    

    def default_package(self):
        # try multiple common/popular C and C++ library include patterns
        if   self.export_include('include', build_dir=True):  pass
        elif self.export_include('include', build_dir=False): pass
        elif self.export_include('src',     build_dir=False): pass
        elif self.export_include('',        build_dir=False): pass

        # default export from {build_dir}/{cmake_build_type}
        if self.export_libs(self.cmake_build_type, src_dir=False): pass
        elif self.export_libs('lib', src_dir=False): pass
    

    ###
    # Perform test steps here with test args
    # `mama test arg1 arg2 arg3`
    # Ex:
    #   def test(self, args): 
    #      self.gdb(f'./RppTests {args}')
    #
    def test(self, args):
        pass



    ############################################


    def cmake_install(self):
        console('\n\n#############################################################')
        console(f"CMake install {self.name} ...")
        run_cmake_build(self)


    def clean_target(self):
        self.dep.clean()


    def ensure_cmakelists_exists(self):
        cmakelists = os.path.join(self.dep.src_dir, 'CMakeLists.txt')
        if not os.path.exists(cmakelists):
            raise IOError(f'Could not find {cmakelists}! Add a CMakelists.txt, or \
                            add `self.nothing_to_build()` to configuration step. \
                            Also note that filename CMakeLists.txt is case sensitive.')
    

    def cmake_build(self):
        console('\n\n#############################################################')
        console(f"CMake build {self.name}")
        self.ensure_cmakelists_exists()
        def cmake_flags():
            flags = ''
            options = self.cmake_opts + cmake_default_options(self) + self.get_product_defines()
            for opt in options: flags += '-D'+opt+' '
            return flags

        self.inject_env()
        run_cmake_config(self, cmake_flags())
        run_cmake_build(self, cmake_buildsys_flags(self))
        self.dep.save_git_status()


    ## Build only this target
    def execute_tasks(self):
        if self.dep.already_executed:
            return
        
        try:
            self.dep.already_executed = True

            if self.dep.should_rebuild and not self.dep.nothing_to_build:
                self.configure() # user customization
                if not self.dep.nothing_to_build:
                    self.build() # user customization
        
            self.package() # user customization

            # no packaging provided by user; use default packaging instead
            if not self.exported_includes or not self.exported_libs:
                self.default_package()

            self.dep.save_exports_as_dependencies(self.exported_libs)

            self.print_exports()

            if self.config.test:
                test_args = self.config.test.lstrip()
                self.test(test_args)
        except:
            console(f'  [BUILD FAILED]  {self.dep.name}')
            raise


    def _print_ws_path(self, what, path):
        n = len(self.config.workspaces_root) + 1
        exists = '' if os.path.exists(path) else '   !! (path does not exist) !!' 
        console(f'    {what}  {path[n:]}{exists}')

    ##
    # Prints out all the exported products
    def print_exports(self):
        console(f'  - Package {self.name}')
        for include in self.exported_includes: self._print_ws_path('<I>', include)
        for library in self.exported_libs:     self._print_ws_path('[L]', library)


######################################################################################


        


from setuptools import setup
import sys

if sys.version_info < (3, 6):
    raise RuntimeError("This package requires Python 3.6+")

setup(name='mama',
      version='0.2.1a31',
      description='A modular C++ build tool even your mama can use',
      url='https://github.com/RedFox20/Mama',
      author='Jorma Rebane',
      author_email='jorma.rebane@gmail.com',
      license='MIT',
      packages=['mama'],
      entry_points = { 'console_scripts': ['mama=mama.main:main'], },
      zip_safe=False,
      python_requires='>=3.6',
      keywords=['mama', 'build', 'mamabuild', 'c', 'c++', 'tool', 'cmake', 'simple', 'easy', 'cross-platform'],
      classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: MIT License',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
      ],
)



import sys, os

## Always flush to properly support Jenkins
def console(s): print(s, flush=True)

def execute(command, echo=False):
    if echo: console(command)
    retcode = os.system(command)
    if retcode != 0:
        raise Exception(f'{command} failed with return code {retcode}')

is_windows = sys.platform == 'win32'
is_linux   = sys.platform.startswith('linux')
is_macos   = sys.platform == 'darwin'
if not (is_windows or is_linux or is_macos):
    raise RuntimeError(f'MamaBuild unsupported platform {sys.platform}')

class System:
    windows = is_windows
    linux   = is_linux
    macos   = is_macos
